import os
import zig
import builtins
import platform
from importlib.util import find_spec
from ctypes import CDLL, POINTER, c_char, c_char_p, c_int, create_string_buffer
from inspect import signature, Parameter
from typing import Set, Callable, Any


def _annotation_to_c_type(type_annotation): 
    match type_annotation:
        case builtins.int:
            return c_int
        case builtins.str:
            return c_char_p
        case _:
            raise Exception(f'Type annotation {param.annotation} not supported')


class ZigLib:
    lib: CDLL

    def __init__(self, lib_name: str):

        match platform.system():
            case 'Darwin':
                name = f'lib{lib_name}.dylib'
            case 'Windows':
                name = f'{lib_name}.dll'
            case _:
                name = f'lib{lib_name}.so'
        
        self.lib = CDLL(os.path.join(os.path.dirname(find_spec('zig').origin), 'zig-out', 'lib', name))
        self.initialized = set()

    def initialize(self, func: Callable) -> None:
        lib_fn = getattr(self.lib, func.__name__)
        sig = signature(func)
        
        lib_fn.restype = _annotation_to_c_type(sig.return_annotation)
        lib_fn.argtypes = [_annotation_to_c_type(x.annotation) for x in sig.parameters.values()]


def _map_arg(arg):
    match arg:
        case int():
            return arg
        case str():
            return create_string_buffer(arg.encode('utf-8'))
        case _:
            raise Exception(f'Runtime argument type {type(arg)} not supported')


def zig_function(zig_lib: ZigLib):
    def decorator(func):

        zig_lib.initialize(func)

        def wrapper(*args):
            args = [_map_arg(x) for x in args]
            fun = getattr(zig_lib.lib, func.__name__)
            res = fun(*args)
            
            match res:
                case bytes():
                    return res.decode('utf-8')
                case int():
                    return res
                case _:
                    raise Exception(f'Return type {type(res)} not supported')

        return wrapper

    return decorator



# Generated from generate_bindings.py
from aber.ziglib import ZigLib, zig_function

lib = ZigLib('hello')

@zig_function(lib)
def zig_add(a: int, b: int) -> int:
    pass

@zig_function(lib)
def zig_mult(a: int, b: int) -> int:
    pass

@zig_function(lib)
def zig_hello(str: str) -> str:
    pass




import os
import sys
import shutil
import base64
import hashlib
import tempfile
import subprocess
import poetry.core.masonry.api as pb
from zipfile import ZipFile
from typing import Any

def read_value(key, lines):
    for line in lines:
        if line.startswith(key):
            return line.split(':')[1].strip()

def sha256_file_b64_nopad(file_path: str) -> str:
    sha256_hash = hashlib.sha256()
    with open(file_path, 'rb') as f:
        while chunk := f.read(8192):
            sha256_hash.update(chunk)
    
    hash_bytes = sha256_hash.digest()
    return base64.urlsafe_b64encode(hash_bytes).rstrip(b'=').decode('utf-8')

def build_wheel(
        wheel_directory: str, 
        config_settings: dict[str, Any] | None = None, 
        metadata_directory: str | None = None
) -> str:

    res = pb.build_wheel(wheel_directory, config_settings, metadata_directory)
    name, version, *_ = res.split('-')

    with tempfile.TemporaryDirectory() as temp_dir:
        wheel = os.path.join(wheel_directory, res)
        with ZipFile(wheel, 'r') as zf:
            zf.extractall(temp_dir)
        
        os.remove(wheel)

        subprocess.call([
            sys.executable,
            '-m',
            'ziglang',
            'build',
        ], cwd=os.path.join(temp_dir, 'zig'))

        shutil.rmtree(os.path.join(temp_dir, 'zig', '.zig-cache'))

        print(os.listdir(os.path.join(temp_dir, 'zig', 'zig-out')))

        lib_dir = os.path.join(temp_dir, 'zig', 'zig-out', 'lib')
        hashes = []
        for x in os.listdir(lib_dir):
            tgt = os.path.join(lib_dir, x)
            hashes.append(f'{os.path.join(name, x)},sha256={sha256_file_b64_nopad(tgt)},{os.path.getsize(tgt)}{os.linesep}')

        with open(os.path.join(temp_dir, f'{name}-{version}.dist-info', 'RECORD'), 'a') as f:
            f.writelines(hashes)

        shutil.make_archive(wheel, 'zip', temp_dir)
        shutil.move(f'{wheel}.zip', wheel)

    return res


build_sdist = pb.build_sdist
get_requires_for_build_sdist = pb.get_requires_for_build_sdist
get_requires_for_build_wheel = pb.get_requires_for_build_wheel
prepare_metadata_for_build_wheel = pb.prepare_metadata_for_build_wheel
